import os
import sys

import numpy as np


def readMesh(file):
    A = np.loadtxt(file, usecols=(1, 2, 3))
    return A


def readOFScal(file, nCells, nHeader=None):
    # Check that the field is not a uniform field
    try:
        f = open(file, "r")
        for i in range(20):
            line = f.readline()
        f.close()
        lineList = line.split()
        if len(lineList) == 3 and lineList[1] == "uniform":
            # Uniform field
            val = float(lineList[2][:-1])
            Array = val * np.ones(nCells)
        else:
            # Not a uniform field
            f = open(file, "r")
            if nHeader is None:
                # Find header
                lines = f.readlines()
                for iline, line in enumerate(lines[:-2]):
                    if str(nCells) in lines[iline] and "(" in lines[iline + 1]:
                        break
                nHeader = iline + 2
                f.close()
            Array = np.loadtxt(file, skiprows=nHeader, max_rows=nCells)
    except:
        print("Issue when reading %s" % file)
        sys.exit()
    return Array


def ofvec2arr(vec):
    vec_list = vec[1:-1].split()
    vec_float = [float(entry) for entry in vec_list]
    return np.array(vec_float)


def readOFVec(file, nCells, nHeader=None):
    # Check that the field is not a uniform field
    try:
        f = open(file, "r")
        for i in range(20):
            line = f.readline()
        f.close()
        lineList = line.split()
        if len(lineList) == 3 and lineList[1] == "uniform":
            # Uniform field
            raise NotImplementedError
            val = ofvec2arr(lineList[2][:-1])
            Array = val * np.ones(nCells)
        else:
            # Not a uniform field
            f = open(file, "r")
            if nHeader is None:
                # Find header
                lines = f.readlines()
                for iline, line in enumerate(lines[:-2]):
                    if str(nCells) in lines[iline] and "(" in lines[iline + 1]:
                        break
                nHeader = iline + 2
                f.close()
            Array = np.loadtxt(
                file, dtype=tuple, skiprows=nHeader, max_rows=nCells
            )
            for i in range(nCells):
                Array[i, 0] = float(Array[i, 0][1:])
                Array[i, 1] = float(Array[i, 1])
                Array[i, 2] = float(Array[i, 2][:-1])
            Array = np.array(Array).astype(float)
    except:
        print("Issue when reading %s" % file)
        sys.exit()

    return Array


def readSizeGroups(file):
    sizeGroup = {}
    f = open(file, "r")
    lines = f.readlines()
    f.close()
    begin = None
    for iline, line in enumerate(lines):
        if "sizeGroups" in line:
            begin = iline + 2
        if not begin is None and ");" in line:
            end = iline - 1
            break
    for line in lines[begin : end + 1]:
        tmp = line.split("{")
        name = tmp[0].strip()
        size = tmp[1].split(";")[0].split()[1]
        sizeGroup[name] = float(size)
    # Sort by size
    sizeGroup = dict(sorted(sizeGroup.items(), key=lambda item: item[1]))
    binGroup = {}
    groups = list(sizeGroup.keys())
    for igroup, group in enumerate(groups):
        if igroup == 0:
            bin_size = (
                sizeGroup[groups[igroup + 1]] - sizeGroup[groups[igroup]]
            )
        elif igroup == len(groups) - 1:
            bin_size = (
                sizeGroup[groups[igroup]] - sizeGroup[groups[igroup - 1]]
            )
        else:
            bin_size_p = (
                sizeGroup[groups[igroup + 1]] - sizeGroup[groups[igroup]]
            )
            bin_size_m = (
                sizeGroup[groups[igroup]] - sizeGroup[groups[igroup - 1]]
            )
            assert abs(bin_size_p - bin_size_m) < 1e-12
            bin_size = bin_size_m
        binGroup[group] = bin_size
    return sizeGroup, binGroup


def getCaseTimes(casePath):
    # Read Time
    times_tmp = os.listdir(casePath)
    # remove non floats
    for i, entry in reversed(list(enumerate(times_tmp))):
        try:
            a = float(entry)
        except ValueError:
            a = times_tmp.pop(i)
            # print('removed ', a)
    time_float = [float(entry) for entry in times_tmp]
    time_str = [entry for entry in times_tmp]
    index_sort = np.argsort(time_float)
    time_float_sorted = [time_float[i] for i in list(index_sort)]
    time_str_sorted = [time_str[i] for i in list(index_sort)]

    return time_float_sorted, time_str_sorted


def getMeshTime(casePath):
    files_tmp = os.listdir(casePath)
    for entry in files_tmp:
        if entry.startswith("meshFaceCentres"):
            return entry[16:-4]



import numpy as np

from bird.utilities.ofio import *


def readFromDict(val_dict, key, read_func=None, path=None, nCells=None):
    if key not in val_dict:
        field = read_func(path, nCells)
        val_dict[key] = field
    else:
        field = val_dict[key]
    return field, val_dict


def check_indLiq(ind_liq, cellCentres):
    height_liq = cellCentres[ind_liq, 1]
    ind_to_remove = np.argwhere(height_liq > 9.5)
    if len(ind_to_remove) > 0:
        ind_liq_copy = ind_liq.copy()
        n_remove = len(ind_to_remove)
        print(f"ind liq found to be at high heights {n_remove} times")
        ind_to_remove = list(ind_liq[ind_to_remove[:, 0]][:, 0])
        ind_liq_copy = list(set(list(ind_liq[:, 0])) - set(ind_to_remove))
        assert len(ind_liq_copy) == len(ind_liq) - n_remove
        ind_liq = np.reshape(np.array(ind_liq_copy), (-1, 1))
    return ind_liq


def check_indHeight(ind_height, cellCentres):
    height_liq = cellCentres[ind_height, 1]
    ind_to_remove = np.argwhere(height_liq < 6)
    if len(ind_to_remove) > 0:
        ind_height_copy = ind_height.copy()
        n_remove = len(ind_to_remove)
        print(f"ind height found to be at low heights {n_remove} times")
        ind_to_remove = list(ind_height[ind_to_remove[:, 0]][:, 0])
        ind_height_copy = list(
            set(list(ind_height_copy[:, 0])) - set(ind_to_remove)
        )
        assert len(ind_height_copy) == len(ind_height) - n_remove
        ind_height = np.reshape(np.array(ind_height_copy), (-1, 1))
    return ind_height


def indLiqFromDict(val_dict, localFolder, nCells, cellCentres):
    if "ind_liq" not in val_dict:
        alpha_gas, val_dict = readFromDict(
            val_dict=val_dict,
            key="alpha_gas",
            read_func=readOFScal,
            path=os.path.join(localFolder, "alpha.gas"),
            nCells=nCells,
        )
        ind_liq = np.argwhere(alpha_gas < 0.8)[:, 0]
        ind_liq = check_indLiq(ind_liq, cellCentres)
        val_dict["ind_liq"] = ind_liq
    else:
        ind_liq = val_dict["ind_liq"]

    return ind_liq, val_dict


def computeGH(localFolder, localFolder_vol, nCells, cellCentres, val_dict={}):
    alpha_gas, val_dict = readFromDict(
        val_dict=val_dict,
        key="alpha_gas",
        read_func=readOFScal,
        path=os.path.join(localFolder, "alpha.gas"),
        nCells=nCells,
    )
    volume, val_dict = readFromDict(
        val_dict=val_dict,
        key="volume",
        read_func=readOFScal,
        path=os.path.join(localFolder_vol, "V"),
        nCells=nCells,
    )
    ind_liq, val_dict = indLiqFromDict(
        val_dict, localFolder, nCells, cellCentres
    )

    holdup = np.sum(alpha_gas[ind_liq] * volume[ind_liq]) / np.sum(
        volume[ind_liq]
    )
    return holdup, val_dict


def computeGH_height(
    localFolder, nCells, cellCentres, height_liq_base, val_dict={}
):
    alpha_gas, val_dict = readFromDict(
        val_dict=val_dict,
        key="alpha_gas",
        read_func=readOFScal,
        path=os.path.join(localFolder, "alpha.gas"),
        nCells=nCells,
    )

    tol = 1e-3
    tol_max = 0.1
    nFound = 0
    iteration = 0
    while nFound <= 10 and tol < tol_max:
        ind_height = np.argwhere(abs(alpha_gas - 0.8) < tol)
        ind_height = check_indHeight(ind_height, cellCentres)
        nFound = len(ind_height)
        tol *= 1.1
        tol = np.clip(tol, a_min=None, a_max=0.2)
        iteration += 1

    if iteration > 1:
        print(f"\tChanged GH tol to {tol:.2g}")

    height_liq = np.mean(cellCentres[ind_height, 1])
    holdup = height_liq / height_liq_base - 1

    return holdup, val_dict


def computeDiam(localFolder, nCells, cellCentres, val_dict={}):
    d_gas, val_dict = readFromDict(
        val_dict=val_dict,
        key="d_gas",
        read_func=readOFScal,
        path=os.path.join(localFolder, "d.gas"),
        nCells=nCells,
    )
    ind_liq, val_dict = indLiqFromDict(
        val_dict, localFolder, nCells, cellCentres
    )

    diam = np.mean(d_gas[ind_liq])

    return diam, val_dict


def computeSpec_liq(
    localFolder, nCells, field_name, key, cellCentres, val_dict={}
):
    species, val_dict = readFromDict(
        val_dict=val_dict,
        key=key,
        read_func=readOFScal,
        path=os.path.join(localFolder, field_name),
        nCells=nCells,
    )
    ind_liq, val_dict = indLiqFromDict(
        val_dict, localFolder, nCells, cellCentres
    )

    species = np.mean(species[ind_liq])

    return species, val_dict


def computeSpec_kla_field(
    localFolder, nCells, key_suffix, cellCentres, val_dict={}, diff=None
):
    if "slip_vel" not in val_dict:
        u_gas, val_dict = readFromDict(
            val_dict=val_dict,
            key="u_gas",
            read_func=readOFVec,
            path=os.path.join(localFolder, "U.gas"),
            nCells=nCells,
        )
        u_liq, val_dict = readFromDict(
            val_dict=val_dict,
            key="u_liq",
            read_func=readOFVec,
            path=os.path.join(localFolder, "U.liquid"),
            nCells=nCells,
        )
        slipvel = np.linalg.norm(u_liq - u_gas, axis=1)
        val_dict["slipvel"] = slipvel

    rho_gas, val_dict = readFromDict(
        val_dict=val_dict,
        key="rho_gas",
        read_func=readOFScal,
        path=os.path.join(localFolder, "thermo:rho.gas"),
        nCells=nCells,
    )
    if "D_" + key_suffix not in val_dict:
        if diff is None:
            T_gas, val_dict = readFromDict(
                val_dict=val_dict,
                key="T_gas",
                read_func=readOFScal,
                path=os.path.join(localFolder, "T.gas"),
                nCells=nCells,
            )
            mu = 1.67212e-06 * np.sqrt(T_gas) / (1 + 170.672 / T_gas)
            D = mu / rho_gas / 0.7
        else:
            D = np.ones(rho_gas.shape) * diff
        val_dict["D_" + key_suffix] = D
    else:
        D = val_dict["D_" + key_suffix]

    d_gas, val_dict = readFromDict(
        val_dict=val_dict,
        key="d_gas",
        read_func=readOFScal,
        path=os.path.join(localFolder, "d.gas"),
        nCells=nCells,
    )
    if "Sh_" + key_suffix not in val_dict:
        # Sh = 1.12*np.sqrt(rho_gas*slipvel*d_gas/(D*0.7*rho_gas))*np.sqrt(0.7)
        Sh = (
            2.0
            + 0.552
            * np.sqrt(rho_gas * slipvel * d_gas / (D * 0.7 * rho_gas))
            * 0.8889
        )
        val_dict["Sh_" + key_suffix] = Sh
    else:
        Sh = val_dict["Sh_" + key_suffix]

    alpha_gas, val_dict = readFromDict(
        val_dict=val_dict,
        key="alpha_gas",
        read_func=readOFScal,
        path=os.path.join(localFolder, "alpha.gas"),
        nCells=nCells,
    )

    kla = Sh * 6 * alpha_gas / d_gas / d_gas * D

    return kla, val_dict


def computeSpec_kla(
    localFolder, nCells, key_suffix, cellCentres, val_dict={}, diff=None
):
    kla, val_dict = computeSpec_kla_field(
        localFolder, nCells, key_suffix, cellCentres, val_dict, diff
    )

    ind_liq, val_dict = indLiqFromDict(
        val_dict, localFolder, nCells, cellCentres
    )

    return np.mean(kla[ind_liq]), val_dict



import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "bird", "requirements.txt")) as f:
    install_requires = f.readlines()

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "bird", "version.py"), encoding="utf-8") as f:
    version = f.read()
version = version.split("=")[-1].strip().strip('"').strip("'")

setup(
    name="nrel-bird",
    version=version,
    description="Bio Reactor Design (BiRD): a toolbox to simulate and analyze different designs of bioreactors in OpenFOAM",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/NREL/BioReactorDesign",
    author="Malik Hassanaly",
    license="BSD 3-Clause",
    package_dir={"bird": "bird"},
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    package_data={
        "": [
            "*requirements.txt",
            "*.json",
            "*.yaml",
            "*.csv",
            "data_conditional_mean",
        ]
    },
    include_package_data=True,
    python_requires=">=3.9",
    install_requires=install_requires,
)



import numpy as np


def conditionalAverage(x, y, nbin):
    try:
        assert len(x) == len(y)
    except:
        print("conditional average x and y have different dimension")
        print("dim x = ", len(x))
        print("dim y = ", len(y))
        sys.exit()
    # Bin conditional space
    mag = np.amax(x) - np.amin(x)
    x_bin = np.linspace(
        np.amin(x) - mag / (2 * nbin), np.amax(x) + mag / (2 * nbin), nbin
    )
    weight = np.zeros(nbin)
    weightVal = np.zeros(nbin)
    asum = np.zeros(nbin)
    bsum = np.zeros(nbin)
    avalsum = np.zeros(nbin)
    bvalsum = np.zeros(nbin)
    inds = np.digitize(x, x_bin)

    a = abs(y - x_bin[inds - 1])
    b = abs(y - x_bin[inds])
    c = a + b
    a = a / c
    b = b / c

    for i in range(nbin):
        asum[i] = np.sum(a[np.argwhere(inds == i)])
        bsum[i] = np.sum(b[np.argwhere(inds == i + 1)])
        avalsum[i] = np.sum(
            a[np.argwhere(inds == i)] * y[np.argwhere(inds == i)]
        )
        bvalsum[i] = np.sum(
            b[np.argwhere(inds == i + 1)] * y[np.argwhere(inds == i + 1)]
        )
    weight = asum + bsum
    weightVal = avalsum + bvalsum

    return x_bin, weightVal / (weight)
